#include <common.h>
#include <i2c.h>

#include "../common/eeprom.h" 

static struct __attribute__ ((__packed__)) eeprom {
	u8 magic[2];           /* 0x0000 - Should be 0xA55A */
	u8 dev_id[2];          /* 0x0002 - Should be 0x006b */
	u8 write_cycle[4];     /* 0x0004 */
	u8 sernum[4];          /* 0x0008 */
	u8 hw_rev[2];          /* 0x000C */
	u8 ident[18];          /* 0x000E */
	u8 reserved[42];       /* 0x0020 */
	u8 user_len[2];        /* 0x004A */
	u8 tbd[58];            /* 0x004C */
	u8 mac1[6];            /* 0x0086 */
	u8 ip1[4];             /* 0x008C */
	u8 subnet1[4];         /* 0x0090 */
	u8 gateway1[4];        /* 0x0094 */
	u8 maintanceip1[4];    /* 0x0098 */
	u8 tbd2[114];          /* 0x009C */
	u8 mac2[6];            /* 0x010E */
	u8 ip2[4];             /* 0x0114 */
	u8 subnet2[4];         /* 0x0118 */
	u8 gateway2[4];        /* 0x011C */
	u8 maintanceip2[4];    /* 0x0120 */
} e;

#define MAGIC_FIRST_BYTE     0xA5
#define MAGIC_SECOND_BYTE    0x5A
#define DEVID_FIRST_BYTE     0x00
#define DEVID_SECOND_BYTE    0x6B

/* Set to 1 if we've read EEPROM into memory */
static int has_been_read = 0;

static int check_valid( void )
{
	if (e.magic[0] != MAGIC_FIRST_BYTE || e.magic[1] != MAGIC_SECOND_BYTE) {
		return -1;
	}
	
	if (e.dev_id[0] != DEVID_FIRST_BYTE || e.dev_id[1] != DEVID_SECOND_BYTE) {
		return -1;
	}
	
	return 0;
}

/**
 * read_eeprom - read the EEPROM into memory
 */
static int read_eeprom(void)
{
	int ret;
	
	i2c_set_bus_num(0);

	if (has_been_read)
		return 0;

	ret = i2c_read(CFG_I2C_EEPROM_ADDR, 0, CFG_I2C_EEPROM_ADDR_LEN,
		(void *)&e, sizeof(e));
#ifdef DEBUG
	show_eeprom();
#endif

	has_been_read = (ret == 0) ? 1 : 0;

	return ret;
}

/**
 * mac_read_from_eeprom - read the MAC addresses from EEPROM
 *
 * This function reads the MAC addresses from EEPROM and sets the
 * appropriate environment variables for each one read.
 *
 * The environment variables are only set if they haven't been set already.
 * This ensures that any user-saved variables are never overwritten.
 *
 * This function must be called after relocation.
 */
int mac_read_from_eeprom(void)
{
	unsigned int i;
	char ethaddr[18];
	char enetvar[9];
	u8 *mac;

	if (read_eeprom()) {
		puts("Read failed.\n");
		return -1;
	}

	if (check_valid() != 0) {
		printf("Invalid Magic (%02x %02x %02x %02x)\n", e.magic[0], e.magic[1], e.dev_id[0], e.dev_id[1]);
		return -1;
	}

	for (i = 0; i < 2; i++) {
		if (i == 0) {
			mac = &e.mac1[0];
		} else {
			mac = &e.mac2[0];
		}
		if (memcmp(mac, "\0\0\0\0\0\0", 6) != 0 &&
		    memcmp(mac, "\xFF\xFF\xFF\xFF\xFF\xFF", 6) != 0) {
			sprintf(ethaddr, "%02X:%02X:%02X:%02X:%02X:%02X", mac[0], mac[1], mac[2],
				mac[3], mac[4], mac[5]);
			sprintf(enetvar, i ? "eth%daddr" : "ethaddr", i);
			/* Override standard mac address */
			printf("Set MAC to %s=%s\n", enetvar, ethaddr);
			setenv(enetvar, ethaddr);
		} else {
			printf("Invalid MAC: %02X:%02X:%02X:%02X:%02X:%02X\n", mac[0], mac[1], mac[2],
				mac[3], mac[4], mac[5]);
		}
	}

	return 0;
}

int echocpu8313_get_ip_address(char *name) {
	unsigned int index;
	char ipaddr[32];
	static const char *ipformat="%d.%d.%d.%d";
	
	if (read_eeprom()) {
		puts("Read failed.\n");
		return -1;
	}

	if (check_valid() != 0) {
		printf("Invalid Magic (%02x %02x %02x %02x)\n", e.magic[0], e.magic[1], e.dev_id[0], e.dev_id[1]);
		return -1;
	}

	if (strncmp(name, "TSEC0", 5) == 0) {
		index = 0;
		if ((e.ip1[0] == 0 && e.ip1[1] == 0 && e.ip1[2] == 0 && e.ip1[3] == 0) ||
			(e.ip1[0] == 0xFF && e.ip1[1] == 0xFF && e.ip1[2] == 0xFF && e.ip1[3] == 0xFF)) {
			printf("No IP defined for TSEC0\n");
			return -1;
		}
		sprintf(ipaddr, ipformat, e.ip1[0], e.ip1[1], e.ip1[2], e.ip1[3]);
		printf("Set IP1 %s\n", ipaddr);
		setenv("ipaddr", ipaddr);
		sprintf(ipaddr, ipformat, e.subnet1[0], e.subnet1[1], e.subnet1[2], e.subnet1[3]);
		setenv("netmask", ipaddr);
		if ((e.gateway1[0] != 0 && e.gateway1[0] != 0xFF) || (e.gateway1[1] != 0 &&  e.gateway1[1] != 0xFF)||
			(e.gateway1[2] != 0 && e.gateway1[2] != 0xFF) || (e.gateway1[3] != 0 && e.gateway1[3] != 0xFF)) {
			sprintf(ipaddr, ipformat, e.gateway1[0], e.gateway1[1], e.gateway1[2], e.gateway1[3]);
			setenv("gatewayip", ipaddr);
		}
		sprintf(ipaddr, ipformat, e.maintanceip1[0], e.maintanceip1[1], e.maintanceip1[2], e.maintanceip1[3]);
		setenv("serverip", ipaddr);
	} else if (strncmp(name, "TSEC1", 5) == 0) {
		if ((e.ip2[0] == 0 && e.ip2[1] == 0 && e.ip2[2] == 0 && e.ip2[3] == 0) ||
			(e.ip2[0] == 0xFF && e.ip2[1] == 0xFF && e.ip2[2] == 0xFF && e.ip2[3] == 0xFF)) {
			printf("No IP defined for TSEC1\n");
			return -1;
		}
		sprintf(ipaddr, ipformat, e.ip2[0], e.ip2[1], e.ip2[2], e.ip2[3]);
		printf("Set IP2 %s\n", ipaddr);
		setenv("ipaddr", ipaddr);
		sprintf(ipaddr, ipformat, e.subnet2[0], e.subnet2[1], e.subnet2[2], e.subnet2[3]);
		setenv("netmask", ipaddr);
		if (e.gateway2[0] != 0 || e.gateway2[1] != 0 ||
			e.gateway2[2] != 0 || e.gateway2[3] != 0) {
			sprintf(ipaddr, ipformat, e.gateway1[0], e.gateway1[1], e.gateway1[2], e.gateway1[3]);
			setenv("gatewayip", ipaddr);
		}
		sprintf(ipaddr, ipformat, e.maintanceip2[0], e.maintanceip2[1], e.maintanceip2[2], e.maintanceip2[3]);
		setenv("serverip", ipaddr);
	}

	return 0;
}
