/*
 * Function for programming the FPGA on the EchoCPU Baord.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS for A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */ 

#include <common.h>
#include <mpc83xx.h>
#include <asm/errno.h>

#include "jtag.h"

#define GPIO_FPGA_RESET   0x80000000
#define GPIO_FPGA_PROG_N  0x08000000
#define GPIO_FPGA_INIT_N  0x02000000
#define GPIO_FPGA_DONE_N  0x04000000

#define GPIO_FPGA_TDO     0x00000004
#define GPIO_FPGA_TDI     0x00020000
#define GPIO_FPGA_TMS     0x00040000
#define GPIO_FPGA_TCK     0x00000002

/**
 * send a TMS clock toggle to the bus
 *
 * @param drv_data: Pointer to driver ctrl structure
 * @param tms: State of TMS (0 or 1)
 * @param tdi: State of TDI (0 or 1)
 */
int fpga_jtag_tms_transition( unsigned int tms, unsigned int tdi)
{
	volatile immap_t *immr = (volatile immap_t *)CFG_IMMR;
	volatile gpio83xx_t *gpio_regs;
	u32 value;
	
	gpio_regs = (volatile gpio83xx_t *)&immr->gpio[0];

	value = gpio_regs->dat;

	/* Set the desired TMS state */
	if ( tms ) {
		value |= GPIO_FPGA_TMS;
	} else {
		value &= ~GPIO_FPGA_TMS;
	}
	if ( tdi ) {
		value |= GPIO_FPGA_TDI;
	} else {
		value &= ~GPIO_FPGA_TDI;
	}
	gpio_regs->dat = value;
	
	/* Set TCK pin to high */
	value |= GPIO_FPGA_TCK;
	gpio_regs->dat = value;
	/* Read TDO */
	value = gpio_regs->dat;
	/* Lower TCLK again */
	value &= ~GPIO_FPGA_TCK;
	gpio_regs->dat = value;

	return ((value & GPIO_FPGA_TDO) != 0);
}

void fpga_jtag_pulse_clock( unsigned int usec )
{
	volatile immap_t *immr = (volatile immap_t *)CFG_IMMR;
	volatile gpio83xx_t *gpio_regs;
	u32 value;
	unsigned int i, delay;
	
	gpio_regs = (volatile gpio83xx_t *)&immr->gpio[0];
	
	delay = usec * 10;
	value = gpio_regs->dat;
	value &= ~GPIO_FPGA_TMS;
	
	for (i=0; i<delay; i++) {
		value |= GPIO_FPGA_TCK;
		gpio_regs->dat = value;
		value &= ~GPIO_FPGA_TCK;
		gpio_regs->dat = value;
	}
}

void fpga_setup_fpga_gpio( void )
{
	volatile immap_t *immr = (volatile immap_t *)CFG_IMMR;
	volatile gpio83xx_t *gpio_regs;
	u32 value;
	
	gpio_regs = (volatile gpio83xx_t *)&immr->gpio[0];
	
	/* Write default value before setting direction! */
	value = gpio_regs->dat;
	value &= ~GPIO_FPGA_TCK;
	value |= GPIO_FPGA_PROG_N | GPIO_FPGA_RESET | GPIO_FPGA_TDI | GPIO_FPGA_TMS;
	gpio_regs->dat = value;
	
	/* RESET must be output,
	   PROG_N must be output */
	value = gpio_regs->dir;
	value &= ~(GPIO_FPGA_INIT_N | GPIO_FPGA_DONE_N | GPIO_FPGA_TDO);
	value |= (GPIO_FPGA_RESET | GPIO_FPGA_PROG_N | GPIO_FPGA_TDI | GPIO_FPGA_TMS | GPIO_FPGA_TCK);
	gpio_regs->dir = value;
	
	value = gpio_regs->odr;
	value &= ~(GPIO_FPGA_INIT_N | GPIO_FPGA_DONE_N | GPIO_FPGA_RESET | GPIO_FPGA_PROG_N | GPIO_FPGA_TDO | GPIO_FPGA_TDI | GPIO_FPGA_TMS | GPIO_FPGA_TCK);
	gpio_regs->odr = value;
	
	value = gpio_regs->imr;
	value &= ~(GPIO_FPGA_INIT_N | GPIO_FPGA_DONE_N | GPIO_FPGA_RESET | GPIO_FPGA_PROG_N | GPIO_FPGA_TDO | GPIO_FPGA_TDI | GPIO_FPGA_TMS | GPIO_FPGA_TCK);
	gpio_regs->imr = value;
}

/**
 * Release the two SPI pins as GPIO after program
 */
static void fpga_release_prog_gpio( void )
{
	volatile immap_t *immr = (volatile immap_t *)CFG_IMMR;
	u32 value;
	volatile gpio83xx_t *gpio_regs;
	
	gpio_regs = (volatile gpio83xx_t *)&immr->gpio[0];

	/* Reset GPIO PIn TMS top PCI_INT_C, TDI to PCI_INT_D */
	/* Reset the GPIO pin to SPI */
	value = gpio_regs->dir;
	value &= ~(GPIO_FPGA_TCK | GPIO_FPGA_TMS | GPIO_FPGA_TDI);
	gpio_regs->dir = value;
	
	value = immr->sysconf.sicrl;
	value &= ~(SICRL_SPI_B | SICRL_SPI_C);
	immr->sysconf.sicrl = value;
}

int fpga_prog_fpga( char *data )
{
	volatile immap_t *immr = (volatile immap_t *)CFG_IMMR;
	volatile gpio83xx_t *gpio_regs;
	u32 value;
	unsigned int count;
	
	gpio_regs = (volatile gpio83xx_t *)&immr->gpio[0];

	fpga_setup_fpga_gpio();
	
	/* Check XSVF File */
	if (data[0] != 0x07 || data[1] != 0x00 || data[2] != 0x13 || data[3] != 0x00) {
		puts("Invalid XSVF File\n");
		return 0;
	}

	/* Put FPGA in reset */
	value = gpio_regs->dat;
	value &= ~(GPIO_FPGA_RESET | GPIO_FPGA_PROG_N);
	gpio_regs->dat = value;
	udelay(1);
	value |= GPIO_FPGA_PROG_N;
	gpio_regs->dat = value;
	udelay(5);

	if (jtag_parse_xsvf_file(data) != 0) {
		puts("Error programming FPGA\n");
		return -1;
	}
	
	/* Test the DONE Pin */
	count = 10000000;
	while ((gpio_regs->dat & GPIO_FPGA_DONE_N) == 0 && count > 0) {
		udelay(1);
		count--;
	}
	if ( count == 0 ) {
		puts("FPGA: It does not accept configuration!\n");
	} else {
		puts("FPGA: OK\n");
	}

	fpga_release_prog_gpio();

	/* Get FPGA out of reset */
	value |= GPIO_FPGA_RESET;
	gpio_regs->dat = value;

	return 0;
}
