/******************************************************************************
 * XSVF Player for the MSP Firmware
 *
 * O. Weber
 * Sep. 2005
 * Built with Code Composer: 2.2
 ******************************************************************************/
#include <common.h>

#include "fpga.h"

/******************************************************************************
* XSVF Command Bytes
*******************************************************************************/

/* encodings of xsvf instructions */
#define XCOMPLETE        0
#define XTDOMASK         1
#define XSIR             2
#define XSDR             3
#define XRUNTEST         4
/* Reserved              5 */
/* Reserved              6 */
#define XREPEAT          7
#define XSDRSIZE         8
#define XSDRTDO          9
#define XSETSDRMASKS     10
#define XSDRINC          11
#define XSDRB            12
#define XSDRC            13
#define XSDRE            14
#define XSDRTDOB         15
#define XSDRTDOC         16
#define XSDRTDOE         17
#define XSTATE           18         /* 4.00 */
#define XENDIR           19         /* 4.04 */
#define XENDDR           20         /* 4.04 */
#define XSIR2            21         /* 4.10 */
#define XCOMMENT         22         /* 4.14 */
#define XWAIT            23         /* 5.00 */
/* Insert new commands here */
#define XLASTCMD         24         /* Last command marker */

/* TAP states */
#define XTAPSTATE_RESET     0x00
#define XTAPSTATE_RUNTEST   0x01    /* a.k.a. IDLE */
#define XTAPSTATE_SELECTDR  0x02
#define XTAPSTATE_CAPTUREDR 0x03
#define XTAPSTATE_SHIFTDR   0x04
#define XTAPSTATE_EXIT1DR   0x05
#define XTAPSTATE_PAUSEDR   0x06
#define XTAPSTATE_EXIT2DR   0x07
#define XTAPSTATE_UPDATEDR  0x08
#define XTAPSTATE_IRSTATES  0x09    /* All IR states begin here */
#define XTAPSTATE_SELECTIR  0x09
#define XTAPSTATE_CAPTUREIR 0x0A
#define XTAPSTATE_SHIFTIR   0x0B
#define XTAPSTATE_EXIT1IR   0x0C
#define XTAPSTATE_PAUSEIR   0x0D
#define XTAPSTATE_EXIT2IR   0x0E
#define XTAPSTATE_UPDATEIR  0x0F
#define XTAPSTATE_MASK      0x0F

/* GPIO Pins used for JTAG */
#define JTAG_TDI     0x01
#define JTAG_TDO     0x02
#define JTAG_TDO_SHIFT     1
#define JTAG_TDO2    0x04
#define JTAG_TDO2_SHIFT    2
#define JTAG_TMS     0x08
#define JTAG_TCK     0x10
#define JTAG_OUT_MASK  (JTAG_TDI | JTAG_TMS | JTAG_TCK)
#define JTAG_IN_MASK (JTAG_TDO | JTAG_TDO2)

/* Return codes used by JTAG functions */
#define JTAG_SUCCESS 0
#define JTAG_MISMATCH -1

/* Some variables to control the TAP state machine */
/* Reflects the actual TAP state */
volatile unsigned char actual_tap_state = XTAPSTATE_RESET;
/* Sets the maximum amount of retries when expecting a value
 * This variable is set by the XREPEAT XSVF command */
volatile unsigned char jtag_max_repeat = 0;
/* Sets the time delay the state machine should stay in 
 * the idle state after a read value which does not
 * match.
 * This variable is set by XRUNTEST XSVF command */
volatile unsigned int run_test_wait_time = 0;
/* This variable show which state should be go to
 * after shifting out/in a value in IR register.
 * Default is idle state.
 * Can be set with the XENDIR XSVF command */
volatile unsigned char xendir_state = XTAPSTATE_RUNTEST;
/* This variable show which state should be go to
 * after shifting out/in a value in DR register.
 * Default is idle state.
 * Can be set with the XENDDR XSVF command */
volatile unsigned char xenddr_state = XTAPSTATE_RUNTEST;
/* This variable defines the amount of bits for the
 * next DR operation. Set by the XSDRSIZE command */
volatile unsigned int dr_bit_amount;
/* This variable stores the TDO mask set by the 
 * XTDOMASK XSVF command */
#define JTAG_MASK_SIZE 64
char tdo_mask[JTAG_MASK_SIZE];
/* Store the read value from TDO*/
char tdo_input_value[JTAG_MASK_SIZE];

/**
 * Struct for evaluating the state machine.
 * The struct show the next state depeding on the level
 * of the TMS pin.
 * tms_null_next: Next state which is reached when TMS=0
 * tms_one_next: Next state which is reached when TMS=1
 * There can also be set two flags with this values:
 * JTAG_NEXT_DEFAULT: This is the default way to go if none of the
 *     two states mentioned in this entry is given.
 *     This flag must be set in every entry or it will be assumed
 *     that TMS=1 is the default way.
 * JTAG_NEXT_GREATER: Every state which is greater than the 
 *     state given in this entry can be reached via this transition.
 *     This flag is special for the SELECTDR state, to go further to any
 *     IR state (TMS=1). It should not be used anywhere else.
 */
#define JTAG_NEXT_DEFAULT 0x40
#define JTAG_NEXT_GREATER 0x80
struct jtag_state_machine_entry {
	char tms_null_next;
	char tms_one_next;
};

/**
 * State transitions for the TAP state machine
 */
struct jtag_state_machine_entry jtag_states[] = {
	/* State XTAPSTATE_RESET */
	{ JTAG_NEXT_DEFAULT | XTAPSTATE_RUNTEST, XTAPSTATE_RESET },
	/* State XTAPSTATE_RUNTEST */
	{ XTAPSTATE_RUNTEST, JTAG_NEXT_DEFAULT | XTAPSTATE_SELECTDR },
	/* State XTAPSTATE_SELECTDR */
	{ JTAG_NEXT_DEFAULT | XTAPSTATE_CAPTUREDR, XTAPSTATE_SELECTIR | JTAG_NEXT_GREATER },
	/* State XTAPSTATE_CAPTUREDR */
	{ XTAPSTATE_SHIFTDR, JTAG_NEXT_DEFAULT | XTAPSTATE_EXIT1DR },
	/* State XTAPSTATE_SHIFTDR */
	{ XTAPSTATE_SHIFTDR, JTAG_NEXT_DEFAULT | XTAPSTATE_EXIT1DR },
	/* State XTAPSTATE_EXIT1DR */
	{ XTAPSTATE_PAUSEDR, JTAG_NEXT_DEFAULT | XTAPSTATE_UPDATEDR},
	/* State XTAPSTATE_PAUSEDR */
	{ XTAPSTATE_PAUSEDR, JTAG_NEXT_DEFAULT | XTAPSTATE_EXIT2DR},
	/* State XTAPSTATE_EXIT2DR */
	{ XTAPSTATE_SHIFTDR, JTAG_NEXT_DEFAULT | XTAPSTATE_UPDATEDR},
	/* State XTAPSTATE_UPDATEDR */
	{ XTAPSTATE_RUNTEST, JTAG_NEXT_DEFAULT | XTAPSTATE_SELECTDR},
	/* State XTAPSTATE_SELECTIR */
	{ JTAG_NEXT_DEFAULT | XTAPSTATE_CAPTUREIR, XTAPSTATE_RESET},
	/* State XTAPSTATE_CAPTUREIR */
	{ XTAPSTATE_SHIFTIR, JTAG_NEXT_DEFAULT | XTAPSTATE_EXIT1IR},
	/* State XTAPSTATE_SHIFTIR */
	{ XTAPSTATE_SHIFTIR, JTAG_NEXT_DEFAULT | XTAPSTATE_EXIT1IR},
	/* State XTAPSTATE_EXIT1IR */
	{ XTAPSTATE_PAUSEIR, JTAG_NEXT_DEFAULT | XTAPSTATE_UPDATEIR},
	/* State XTAPSTATE_PAUSEIR */
	{ XTAPSTATE_PAUSEIR, JTAG_NEXT_DEFAULT | XTAPSTATE_EXIT2IR},
	/* State XTAPSTATE_EXIT2IR */
	{ XTAPSTATE_SHIFTIR, JTAG_NEXT_DEFAULT | XTAPSTATE_UPDATEIR},
	/* State XTAPSTATE_UPDATEIR */
	{ XTAPSTATE_RUNTEST, JTAG_NEXT_DEFAULT | XTAPSTATE_SELECTDR}
};

static void waitTime(unsigned int usec)
{
	fpga_jtag_pulse_clock(usec);
}

/**
 * Get a unsigned int (16bit value) from a XSVF data stream
 * 
 * @param source: Pointer to the XSVF data stream
 */
static unsigned int get_2b_value( const char *source)
{
	unsigned int value = 0;
	
	/* Get high byte */
	value |= (*source++) << 8;
	/* Get low byte */
	value |= *source;
	
	return value;
}

/**
 * Go the the given new_state.
 * This function passes all states which must be covered to reach
 * the given end_state.
 * 
 * @param new_state: State to reach
 */
static void gotoTapState(unsigned char new_state)
{
	unsigned int i;
	
	/* Reset goes special */
	if ( new_state == XTAPSTATE_RESET ) {
		/* Toggle the TCL pin for 5 times as this is the longest way to get to reset
		* state (if TAP is already in reset, giving TMS=1 holds it there) */
		for ( i = 0; i < 6; ++i )
		{
			fpga_jtag_tms_transition( 1, 0 );
		}
		/* Reset state reached */
		actual_tap_state = XTAPSTATE_RESET;
		return;
	} else if ( new_state == actual_tap_state) {
		/* Already in target state.  Do nothing except when in DRPAUSE
		* or in IRPAUSE to comply with SVF standard.
		* Then go to the exit2 state. */
		if ( actual_tap_state == XTAPSTATE_PAUSEDR ) {
			fpga_jtag_tms_transition( 1, 0 );
			actual_tap_state = XTAPSTATE_EXIT2DR;
		} else if ( actual_tap_state == XTAPSTATE_PAUSEIR ) {
			fpga_jtag_tms_transition( 1, 0 );
			actual_tap_state = XTAPSTATE_EXIT2IR;
		}
	} else {
		/* Do state transistions as long as the desired state
		* is reached */
		while (new_state != actual_tap_state) {
			/* TMS = 0 is the default way */
			if ((jtag_states[actual_tap_state].tms_null_next & JTAG_NEXT_DEFAULT) != 0) {
				/* Check a state greater then the state reachable via TMS=1 must
				 * be reached. This test is special for the SELECTDR state! */
				if ((jtag_states[actual_tap_state].tms_one_next & JTAG_NEXT_GREATER) != 0) {
					/* If new_state is greater or equal the state given in the transition
					 * struct then go the TMS=1 way */
					if (new_state >= (jtag_states[actual_tap_state].tms_one_next & XTAPSTATE_MASK)) {
						fpga_jtag_tms_transition( 1, 0 );
						actual_tap_state = (jtag_states[actual_tap_state].tms_one_next & XTAPSTATE_MASK);
					} else {
						/* The state can be rached via TMS=0 */
						fpga_jtag_tms_transition( 0, 0 );
						actual_tap_state = (jtag_states[actual_tap_state].tms_null_next & XTAPSTATE_MASK);
					}
				} else {
					/* Only when new_state matches exactly the state given in the state
					 * transition structure then go TMS=1 */
				  if (new_state == (jtag_states[actual_tap_state].tms_one_next & XTAPSTATE_MASK)) {
						fpga_jtag_tms_transition( 1, 0 );
						actual_tap_state = (jtag_states[actual_tap_state].tms_one_next & XTAPSTATE_MASK);
					} else {
						/* No, so go TMS = 0 way */
						fpga_jtag_tms_transition( 0, 0 );
						actual_tap_state = (jtag_states[actual_tap_state].tms_null_next & XTAPSTATE_MASK);
					}
				}
			} else {
				/* If the state following with TMS=0 is the state to be reached, then
				 * go this way. */
			  if (new_state == (jtag_states[actual_tap_state].tms_null_next & XTAPSTATE_MASK)) {
					fpga_jtag_tms_transition( 0, 0 );
					actual_tap_state = (jtag_states[actual_tap_state].tms_null_next & XTAPSTATE_MASK);
				} else {
					/* Go the default way with TMS=1 */
					fpga_jtag_tms_transition( 1, 0 );
					actual_tap_state = (jtag_states[actual_tap_state].tms_one_next & XTAPSTATE_MASK);
				}
			}
		}
	}
}

/**
 * Shift in data in the device and get the value shifted out.
 * If exit shift is set, the function will set TMS=1 with the last bit
 * shifted, so that the TAP goes to EXIT1DR/EXIT1IR state
 * 
 * @param inval: Pointer to a buffer containing the data to be shifted out
 * @param outval: Pointer to a buffer for the data shifted in
 * @param len: Amount of bits to shift
 * @param exitShift: Go to the exit state after all bits were shifted
 */
static void jtag_shift_value(const char *inval, char *outval, unsigned int len, char exitShift)
{
	int i;
	unsigned char shift_byte;
	unsigned int tms = 0, tdi, tdo;
	
	/* Most significant byte comes first!
	 * So it must be stored at the end opf the output buffer! */
	if (outval != NULL) {
		if (len > 7) {
			outval += ((len >> 3) - 1);
			/* If bit length is not a multiple of 8 than add
			 * one byte */
			if ((len & 0x07) != 0) {
				outval++;
			}
		}
	}
	/* Same for the input buffer. The byte first to send
	 * stands at the end in the XSVF stream. So the data
	 * is passed from end to start of buffer.
	 * Add the length in bytes to the buffer. */
	if (len > 7) {
		inval += ((len >> 3) - 1);
		/* If bit length is not a multiple of 8 than add
		 * one byte */
		if ((len & 0x07) != 0) {
			inval++;
		}
	}
	/* Shift out all bits */
	while ( len ) {
		/* Get the byte to shift */
		shift_byte = *inval;
		/* Clear output buffer, if available */
		if ( outval != NULL ) {
			*outval = 0;
		}
		/* Shift the byte out bit by bit or until len bits were shifted.
		* The remaining bits of the last byte are not shfited when len is
		* already 0 */
		for ( i = 0; ( len != 0 && ( i < 8 ) ); i++ ) {
			/* Decrement the bit amount */
			--len;
			/* Is this the last bit and exitShift is set ? */
			if ( exitShift && !len ) {
				/* Yes, so set the TMS pin to high to invoke the state transition 
				* to IR/DREXIT1*/
				tms = 1;
				/* Update TAP state:  Shift->Exit */
				++actual_tap_state;
			}

			/* Set the new TDI level according to the bit value to be shifted out */
			tdi = (shift_byte & 1);
			/* Shift the value one bit to the right */
			shift_byte >>= 1;

			/* Set TCK low */
			tdo = fpga_jtag_tms_transition( tms, tdi );

			/* If the values which are shifted out from the device
			* should be stored, then do this to the output buffer */
			if ( outval != NULL && tdo != 0 ) {
				*outval |= (1 << i);
			}
		}
		/* Switch to the next value in in- and output buffer */
		if (outval != NULL) {
			outval--;
		}
		inval--;
	}
}

/**
 * First go the the given new_state.
 * Thrn shift in data in the device and get the value shifted out.
 * If expected is not 0 the outputted value is checked against
 * the value in expected.
 * 
 * @param new_state: State to go before shifting in/out a value
 * @param end_state: State to go after shifting out a value
 *    If new_value == end_value then the state machine stays in the shift state
 * @param inval: Pointer to a buffer containing the data to be shifted out
 * @param outval: Pointer to a buffer for the data shifted in
 * @param expected: Pointer to a buffer containing the value expected to be shifted in
 * @param len: Amount of bits to shift
 */
static int jtag_go_n_shift_value(char new_state, char end_state,
  const char *inval, char *outval, 
  const char *expected, unsigned int len)
{
	volatile unsigned int i;
	unsigned int tocompare;
	unsigned char exit_shift = ( new_state != end_state );
	unsigned char mismatch = 0;
	unsigned char retry = 0;
	if ( !len ) {
		/* Compatibility with XSVF2.00:  XSDR 0 = no shift, but wait in RTI */
		if ( run_test_wait_time != 0 )
		{
			/* Wait for prespecified XRUNTEST time */
			gotoTapState( XTAPSTATE_RUNTEST );
			waitTime( run_test_wait_time );
		}
	} else {
		do {
			/* Goto Shift-DR or Shift-IR */
			gotoTapState( new_state );

			/* Shift TDI and capture TDO */
			jtag_shift_value( inval, outval, len, exit_shift );

			if ( expected != NULL ) {
				/* Compare TDO data to expected TDO data */
				tocompare = (len >> 3);
				if ((len & 0x07) != 0) {
					tocompare++;
				}
				for (i=0; i<tocompare; i++) {
					if ((outval[i] & tdo_mask[i]) != (expected[i] & tdo_mask[i])) {
						mismatch = 1;
					}
				}
			}

			if ( exit_shift ) {
				if ( mismatch && run_test_wait_time && ( retry < jtag_max_repeat ) ) {
					/* Do exception handling retry - ShiftDR only */
					gotoTapState( XTAPSTATE_PAUSEDR );
					/* Shift 1 extra bit */
					gotoTapState( XTAPSTATE_SHIFTDR );
					/* Increment RUNTEST time by an additional 25% */
					run_test_wait_time += ( run_test_wait_time >> 2 );
				} else {
					/* Do normal exit from DR/IREXIT */
					gotoTapState( end_state );
				}

				if ( run_test_wait_time != 0 ) {
				/* Wait for prespecified XRUNTEST time */
				gotoTapState( XTAPSTATE_RUNTEST );
				waitTime( run_test_wait_time );
				}
			}
		} while ( mismatch && ( retry++ < jtag_max_repeat ) );
	}
	if (mismatch != 0) {
		return JTAG_MISMATCH;
	}
	return JTAG_SUCCESS;
}

/* Implement one function for each XSVF command */

/**
 * Implemement the XREPEAT command:
 * Defines how often the read of a value will be retried
 * when the result does not match the expected result
 * 
 * @param filepos: Pointer to a char pointer to the XSVF data
 */
static int xsvfDoXREPEAT( const char **filepos )
{
	/* Store the max repeat variable */
	jtag_max_repeat = **filepos;
	/* Advance the file buffer */
	(*filepos)++;
	return JTAG_SUCCESS;
}

/**
 * Implemement the XRUNTEST command
 * Defines the time how long to stay in the idle state
 * 
 * @param filepos: Pointer to a char pointer to the XSVF data
 */
static int xsvfDoXRUNTEST( const char **filepos )
{
	/* Store the wait time in the variable */
	run_test_wait_time = (get_2b_value( *filepos ) << 16);
	*filepos += 2;
	run_test_wait_time |= get_2b_value( *filepos );
	/* Advance the file buffer */
	*filepos += 2;
	return JTAG_SUCCESS;
}

/**
 * Implemement the XSIR command
 * Go to the SHIFTIR state and shift in a value
 * 
 * @param filepos: Pointer to a char pointer to the XSVF data
 */
static int xsvfDoXSIR( const char **filepos )
{
	unsigned char shift_bits;
	int error_code;
	unsigned int offset;

	/* Get the shift length and store */
	shift_bits = **filepos;
	(*filepos)++;
  
 	error_code = jtag_go_n_shift_value( XTAPSTATE_SHIFTIR, xendir_state,
                               *filepos, NULL, NULL, shift_bits);
	/* Shift the data */
	offset = (shift_bits >> 3);
	if ((shift_bits & 0x07) != 0) {
		offset++;
	}
	*filepos += offset;

	return error_code;
}

/**
 * Implemement the XSDRSIZE command
 * Defines the amount of bits to be shifted in SHIFTDR state
 * 
 * @param filepos: Pointer to a char pointer to the XSVF data
 */
static int xsvfDoXSDRSIZE( const char **filepos )
{
	/* Store the last two byte in the length field */
	*filepos += 2;
	dr_bit_amount = get_2b_value(*filepos);
	/* Advance the file buffer */  
	*filepos += 2;
	return JTAG_SUCCESS;
}

/**
 * Implemement the XTDOMASK command
 * Defines the mask to be used by the next XSDRTDO command.
 * 
 * @param filepos: Pointer to a char pointer to the XSVF data
 */
static int xsvfDoXTDOMASK( const char **filepos )
{
	unsigned int offset;
	
	offset = (dr_bit_amount >> 3);
	if ((dr_bit_amount & 0x07) != 0) {
		offset++;
	}
	if (offset > JTAG_MASK_SIZE) {
		return JTAG_OUT_MASK;
	}
	/* Copy the mask to the mask buffer */
	memcpy(tdo_mask, *filepos, offset);
	*filepos += offset;

	return JTAG_SUCCESS;
}

/**
 * Implemement the XSDRTDO command
 * Shifts out a value and checks for an expected shifted in value
 * 
 * @param filepos: Pointer to a char pointer to the XSVF data
 */
static int xsvfDoXSDRTDO( const char **filepos )
{
	int error_code;
	unsigned int offset; 
	
	offset = (dr_bit_amount >> 3);
	if ((dr_bit_amount & 0x07) != 0) {
		offset++;
	}
	error_code = jtag_go_n_shift_value( XTAPSTATE_SHIFTDR, xenddr_state,
                               *filepos, tdo_input_value,
                               (*filepos) + offset, dr_bit_amount);

	/* Add the length of two values to the file pointer
	 * 2 * dr_bit_amount / 8 => dr_bit_amount / 4 => dr_bit_amount >> 2 */
	*filepos += dr_bit_amount >> 2;
	return error_code;
}

/**
 *  XSDRB/XSDRC/XSDRE <lenVal.TDI[XSDRSIZE]>
 *  If not already in SHIFTDR, goto SHIFTDR. Shift the given TDI data into the JTAG scan chain.
 *  Ignore TDO. If cmd==XSDRE, then goto ENDDR.  Otherwise, stay in ShiftDR.
 *  XSDRB, XSDRC, and XSDRE are the same implementation.
 *
 * @param filepos: Pointer to a char pointer to the XSVF data
 * @param command: Command byte to test if it was a XSDRE command
 *
 * @return == JTAG_SUCCESS OK or error code 
 */
static int xsvfDoXSDRBCE( const char **filepos, unsigned char command )
{
	unsigned char end_state;
	int error_code;
	
	/* Get the shift length and store */
	end_state = (unsigned char)(( command == XSDRE ) ? xenddr_state : XTAPSTATE_SHIFTDR);
	error_code  = jtag_go_n_shift_value( XTAPSTATE_SHIFTDR, end_state, *filepos, NULL, NULL, dr_bit_amount);
	
	/* Add the length of two values to the file pointer
	 * dr_bit_amount / 8 */
	*filepos += dr_bit_amount >> 3;
	if ((dr_bit_amount & 0x7) != 0) {
		(*filepos)++;
	}
	
	return error_code;
}

/**
 * Implemement the XSTATE command
 * Go to the given state.
 * 
 * @param filepos: Pointer to a char pointer to the XSVF data
 */
int xsvfDoXSTATE( const char **filepos )
{
	gotoTapState( **filepos );
    
	(*filepos)++;
    
	return JTAG_SUCCESS;
}

/**
 * Implemement the XENDIR command
 * Defines the state to go after shift IR.
 * 
 * @param filepos: Pointer to a char pointer to the XSVF data
 */
int xsvfDoXENDIR( const char **filepos )
{ 
	/* Copy the mask to the mask buffer */
	if (**filepos == 0x01) {
		xendir_state = XTAPSTATE_PAUSEIR;
	} else {
		xendir_state = XTAPSTATE_RUNTEST;
	}
	(*filepos)++;

	return JTAG_SUCCESS;
}

/**
 * Implemement the XENDIR command
 * Defines the state to go after shift DR.
 * 
 * @param filepos: Pointer to a char pointer to the XSVF data
 */
int xsvfDoXENDDR( const char **filepos )
{ 
	/* Copy the mask to the mask buffer */
	if (**filepos == 0x01) {
		xendir_state = XTAPSTATE_PAUSEDR;
	} else {
		xendir_state = XTAPSTATE_RUNTEST;
	}
	(*filepos)++;

	return JTAG_SUCCESS;
}

/**
 * Implemement the XWAIT command
 * Go to state A, wait for n muicroseconds and then go the state B
 * 
 * @param filepos: Pointer to a char pointer to the XSVF data
 */
int xsvfDoXWAIT( const char **filepos )
{
	unsigned char   wait_state;
	unsigned char   end_state;
	unsigned int   wait_time;

	/* Get Parameters */
	/* <wait_state> */
	wait_state = **filepos;
	(*filepos)++;

	/* <end_state> */
	end_state = **filepos;
	(*filepos)++;
	
	/* Store the last two byte in the length field */
	wait_time = (get_2b_value(*filepos) << 16);
	*filepos += 2;
	wait_time |= get_2b_value(*filepos);
	/* Advance the file buffer */  
	*filepos += 2;

	/* If not already in <wait_state>, go to <wait_state> */
	if ( actual_tap_state != wait_state )
	{
	gotoTapState( wait_state );
	}

	/* Wait for <wait_time> microseconds */
	waitTime( wait_time );

	/* If not already in <end_state>, go to <end_state> */
	if ( actual_tap_state != end_state )
	{
	gotoTapState( end_state );
	}

	return JTAG_SUCCESS;
}

/**
 * Execute one XSVF command or a complete XSVF file
 * 
 * @param filepos: Pointer to a char pointer to the XSVF data
 * @param single:  = 1 Only execute one command and exit
 *                 = 0 Execute until XCOMPLETE command in XSVF stream
 */
int jtag_parse_xsvf_file(const char *fileptr)
{
	int error_code = 0;
	unsigned char cmd;

	/* Do until XCOMPLETE command (= 0x00) */
	while (*fileptr != 0x00) {
		/* Get function pointer in XSVF command handler array */
		cmd = *fileptr;
		/* Skip command byte */
		fileptr++;
		switch (cmd) {
			case XTDOMASK:
				error_code = xsvfDoXTDOMASK(&fileptr);
				break;
			case XSIR:
				error_code = xsvfDoXSIR(&fileptr);
				break;
			case XRUNTEST:
				error_code = xsvfDoXRUNTEST(&fileptr);
				break;
			case XREPEAT:
				error_code = xsvfDoXREPEAT(&fileptr);
				break;
			case XSDRSIZE:
				error_code = xsvfDoXSDRSIZE(&fileptr);
				break;
			case XSDRTDO:
				error_code = xsvfDoXSDRTDO(&fileptr);
				break;
			case XSDRB:
			case XSDRC:
			case XSDRE:
				error_code = xsvfDoXSDRBCE(&fileptr, cmd);
				break;
			case XSTATE:
				error_code = xsvfDoXSTATE(&fileptr);
				break;
			case XENDIR:
				error_code = xsvfDoXENDIR(&fileptr);
				break;
			case XENDDR:
				error_code = xsvfDoXENDDR(&fileptr);
				break;
			case XWAIT:
				error_code = xsvfDoXWAIT(&fileptr);
				break;
			default:
				/* Else send error code, that command is not available */
				printf("Unknown command 0x%02x at offset 0x%x\n", cmd, (unsigned int)fileptr - 0x400000);
				return -1;
		}
		if (error_code != JTAG_SUCCESS) {
			printf("Command 0x%02x failed with %d\n", cmd, error_code);
			return -1;
		}
	}
	
	return error_code;
}




